;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;	TEMPLATES
;	two templates for the program to use
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(deftemplate player
	(slot name)
	(multislot hand))
	
(deftemplate hit 
	(slot name))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;	FUNCTIONS
;	these hadle card calculations and generation
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(deffunction numerical-value-card ( ?v )
	(if	(or (eq ?v J) (eq ?v Q) (eq ?v K))
		then 10
		else ?v)
	)

(deffunction add-all (?list ?v)
	(bind ?outlist (create$ ))
	(if (neq ?v A)
		then	(progn$ (?c ?list)
				(bind ?outlist (insert$ ?outlist 1 (+ ?c (numerical-value-card ?v) )))
				)
		else	(progn$ (?c ?list)
				(bind ?outlist (insert$ ?outlist 1 (+ ?c 1 )))
				(bind ?outlist (insert$ ?outlist 1 (+ ?c 11 )))
				)
		)
	?outlist
	)
	
(deffunction numerical-value-hand (?vsofar ?v)
	(if (= (length ?v) 1)
		then	(bind ?t (add-all ?vsofar (nth$ 1 ?v)) ) 
		else	(bind ?t (numerical-value-hand (add-all ?vsofar (nth$ 1 ?v)) (rest$ ?v) ))
		)
	?t
	)

(deffunction is-win (?hand)
	(bind ?result FALSE)
 	(if (= (length ?hand) 5)
		then
			(progn$ (?v (numerical-value-hand (create$ 0) ?hand))  	; if the number of cards is 5 and at least one possible value is less than 22 win
				(if (< ?v 22) then (bind ?result TRUE))	; ie: if all possible totals are not losses 
				)
		else
			(progn$ (?v (numerical-value-hand (create$ 0) ?hand))  	; if any possible result is a win
				(if (= ?v 21) then (bind ?result TRUE))	; 
				)
		)
	?result
	)
(deffunction is-loss (?hand)
	(bind ?result TRUE)
 	(progn$ (?v (numerical-value-hand (create$ 0) ?hand))  	; if all possible values are greater than 21
		(if (< ?v 22) then (bind ?result FALSE)) 
		)
	?result
	)
 
(deffunction has-two (?value ?list)
	(bind ?t 0)
	(progn$ (?c ?list)
		(if (eq ?c ?value) then (bind ?t (+ ?t 1)))
		)
	(>= ?t 2)
	)
	
(deffunction random-card-value ()
	(bind ?r (random 2 14))
		(if (= ?r 11) then (bind ?r J)
		else (if (= ?r 12) then (bind ?r Q)
		else (if (= ?r 13) then (bind ?r K)
		else (if (= ?r 14) then (bind ?r A)))))
	?r	
	)
(deffunction new-random-card (?cardsinplay)
	(if (= (length ?cardsinplay) 0) then
		(while (= 1 1)
			(bind ?r (random-card-value))
			(if (not (has-two ?r ?cardsinplay)) then (break))
			)
		else
			(bind ?r (random-card-value))
		)
	?r
	)
	
(deffunction assert-start ()
	(bind ?p1cards (create$ ))
	(bind ?p2cards (create$ ))
	
	(bind ?p1cards (insert$ ?p1cards 1 (new-random-card (create$))))
	(bind ?p1cards (insert$ ?p1cards 1 (new-random-card ?p1cards)))
	(bind ?p2cards (insert$ ?p2cards 1 (new-random-card ?p1cards)))
	(bind ?p2cards (insert$ ?p2cards 1 (new-random-card (insert$ ?p1cards 1 ?p2cards))))
	
	(assert (player (name dealer) (hand ?p1cards)))
	(assert (player (name player) (hand ?p2cards)))
	TRUE
	)
	
(deffunction minvalue (?list)
	(bind ?cmin (nth$ 1 ?list))
	(progn$ (?v ?list)
		(if (<= ?v ?cmin) then (bind ?cmin ?v))
		)
	?cmin
	)
	
(deffunction compare-hands (?hand1 ?hand2)
	(bind ?res 0)
	(if (< (minvalue (numerical-value-hand (create$ 0) ?hand1)) (minvalue (numerical-value-hand (create$ 0) ?hand2))) 
		then (bind ?res -1) 
		else (if (> (minvalue (numerical-value-hand (create$ 0) ?hand1)) (minvalue (numerical-value-hand (create$ 0) ?hand2))) 
			then (bind ?res 1)
			else (bind ?res 0)))
	?res
)
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;	RULES
;	these rules handle the game (hits, wins and initial values)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defrule start
	(initial-fact)
	=>
	(assert-start)
	)
	
(defrule loser (declare (salience 99))
	(not (stop))
	(player (name ?n1) (hand $?h1))
	(player (name ?n2) (hand $?h2))
	(test (neq ?n1 ?n2))
	(test (or (is-loss (create$ ?h1)) (is-loss (create$ ?h2))))
	=>
	(open "dribble.txt" file "a")
	(if (is-loss (create$ ?h1))
		then (printout file ?n2 crlf)
		else (printout file ?n1 crlf))
	(close file)
	(assert (stop))
	)
(defrule winner (declare (salience -1))
	(not (stop))
	(not (hit))
	(player (name ?n1) (hand $?h1))
	(player (name ?n2) (hand $?h2))
	(test (neq ?n1 ?n2))
	=>
	(open "dribble.txt" file "a")
	(if (and (is-win (create$ ?h1)) (is-win (create$ ?h2)))
		then	(printout file "draw" crlf)
		else	(if (is-win (create$ ?h1))
				then (printout file ?n1 crlf)
				else (if (is-win (create$ ?h2))
					then (printout file ?n2 crlf)
					else (if (= -1 (compare-hands (create$ ?h2) (create$ ?h1)))
						then (printout file ?n2 crlf)
						else (if (= -1 (compare-hands (create$ ?h1) (create$ ?h2)))
							then (printout file ?n1 crlf)
							else (printout file "draw" crlf))))))
	(close file)
	(assert (stop))
	)

(defrule hit (declare (salience 100))
	(not (stop))
	?hitfact <- (hit (name ?n))
	?playerfact <- (player (name ?n) (hand $?h1))
	(player (name ~?n) (hand $?h2))
	=>
	(retract ?hitfact)
	(modify ?playerfact (hand ?h1 (new-random-card (insert$ ?h1 1 ?h2))))
	)
	

	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;	YOUR CODE
;	any rules functions etc. you need to make the hit rule, and most importantly, your hit rule
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defrule dealer-hit (declare (salience 2))
	(player (name dealer) (hand $?h))
	(test (not (member$ 17 (numerical-value-hand (create$ 0) (create$ ?h)))))
	=>
	(assert (hit (name dealer)))
	)
	
(defrule make-hit (declare (salience 3))
	;;; PUT YOUR POLICY HERE
	=>
	(assert (hit (name player)))
	)

; extra notes on function and templates
; all the players are in templates of the form (player (name dealer) (hand A Q 2)), the only players are "dealer" and "player"
; to make a hit (assert (hit (name foo)))

; functions that might be useful (and how to call them):
;	(is-win (create$ ?hand))	true if the hand is a win, false otherwise
;	(is-loss (create$ ?hand))	true if the hand is a losing hand, false otherwise
;	note that a hand can be niether winning nor losing (those that play can continue on)
;
;	(numerical-value-hand (create$ 0) (create$ ?h1))	gives the total of the hand value (a list of possible values: to take into account aces)
; 	(compare-hand (create$ ?h1) (create$ ?h2))		-1 if hand 1 is less than hand 2, 1 if hand 1 is greater than hand 2 and 0 if the hands are the same value

	